use std::{any::TypeId, sync::Arc};

use ahash::AHashMap;
use parking_lot::RwLock;
use tokio::sync::mpsc;

use crate::{
    event::{internal::EventResolver, Event},
    on,
    system::{
        internal::{EventHandlerHandle, SystemHandle},
        System,
    },
    Dependency, SystemRef,
};

pub struct Context {
    systems: Arc<RwLock<Vec<SystemHandle>>>,

    sender: mpsc::UnboundedSender<Event>,
    receiver: mpsc::UnboundedReceiver<Event>,

    handlers: AHashMap<TypeId, EventResolver>,
}

impl Default for Context {
    fn default() -> Self {
        Self::new()
    }
}

impl Context {
    pub fn new() -> Self {
        let (sender, receiver) = mpsc::unbounded_channel();

        Self {
            systems: Arc::new(RwLock::new(Vec::new())),
            sender,
            receiver,
            handlers: AHashMap::new(),
        }
    }

    pub fn add<S: System>(&mut self) {
        self.get_or_init_dependency(&Dependency::new::<S>());
    }

    pub async fn start(mut self) {
        while let Some(event) = self.receiver.recv().await {
            self.dispatch_event(event);
        }
    }

    fn stop(&mut self) {
        let mut systems = self.systems.write();
        for i in (0..systems.len()).rev() {
            systems.remove(i);
        }
    }

    fn dispatch_event(&mut self, event: Event) {
        if let Some(handler) = self.handlers.get(&event.id) {
            handler.dispatch_event(event);
        } else if event.id == TypeId::of::<on::Exit>() {
            self.receiver.close();
            self.stop();
        }
    }

    fn get_or_init_dependency(&mut self, dependency: &Dependency) -> usize {
        match self.get_dependency(dependency) {
            Some(index) => index,
            None => self.init_dependency(dependency),
        }
    }

    fn init_dependency(&mut self, dependency: &Dependency) -> usize {
        let dependencies = dependency.dependencies;
        let mut dependency_ids = vec![];
        for dependency in dependencies {
            dependency_ids.push((
                dependency.id.into(),
                self.get_or_init_dependency(dependency),
            ));
        }

        let event_sender = self.sender.clone();

        let context_handle = ContextHandle {
            systems: self.systems.clone(),
            dependency_ids,
            event_sender,
        };

        let builder = (dependency.builder)();
        let system_handle = builder.create(context_handle);

        let index = self.systems.read().len();

        let event_handlers = dependency.event_handlers;
        for event_handler in event_handlers {
            // TODO: Add event handler
            let event_handler_builder = (event_handler.builder)();
            let event_handler = event_handler_builder.create(event_handler.handler_any);
            let event_handler_handle =
                EventHandlerHandle::new(system_handle.system.clone(), event_handler);
            self.register_handler(event_handler_handle);
        }

        self.systems.write().push(system_handle);

        index
    }

    fn get_dependency(&self, dependency: &Dependency) -> Option<usize> {
        let id = dependency.id.into();
        let systems = &self.systems.read();

        (0..systems.len()).find(|&i| systems[i].id == id)
    }

    fn register_handler(&mut self, handler: EventHandlerHandle) {
        let id = handler.event_id;
        if !self.handlers.contains_key(&id) {
            self.handlers.insert(id, EventResolver::new());
        }

        match self.handlers.get_mut(&id) {
            Some(resolver) => resolver.register_handler(handler),
            None => panic!("EventResolver disappeared"),
        }
    }
}

#[derive(Clone)]
pub struct ContextHandle {
    systems: Arc<RwLock<Vec<SystemHandle>>>,
    dependency_ids: Vec<(TypeId, usize)>,
    event_sender: mpsc::UnboundedSender<Event>,
}

impl ContextHandle {
    pub fn obtain<S: System>(&self) -> SystemRef<S> {
        let index = self.dependency_index::<S>();
        let systems = self.systems.read();
        let system = systems[index].system.clone();

        SystemRef::new(system)
    }

    pub fn raise<T: 'static + Send + Sync>(&self, event: T) {
        let event = Event::new(event);
        let _ = self.event_sender.send(event);
    }

    fn dependency_index<S: System>(&self) -> usize {
        let id = TypeId::of::<S>();
        for dependency in &self.dependency_ids {
            if dependency.0 == id {
                return dependency.1;
            }
        }
        panic!("Requested unavailable or invalid dependency");
    }
}
