use std::{
    any::{Any, TypeId},
    marker::PhantomData,
    sync::Arc,
};

use parking_lot::RwLock;

use crate::{context::ContextHandle, System};

pub struct SystemHandle {
    pub(crate) system: Arc<RwLock<dyn Any + Send + Sync>>,
    pub(crate) id: TypeId,
}

pub trait SystemBuilder {
    fn create(&self, ctx: ContextHandle) -> SystemHandle;
}

pub struct TypedSystemBuilder<S: System> {
    marker: PhantomData<S>,
}

impl<S: System> TypedSystemBuilder<S> {
    pub fn get_builder() -> Box<dyn SystemBuilder> {
        Box::new(Self {
            marker: PhantomData,
        })
    }
}

impl<S: System> SystemBuilder for TypedSystemBuilder<S> {
    fn create(&self, ctx: ContextHandle) -> SystemHandle {
        let mut system = S::new(ctx);

        system.start();

        SystemHandle {
            system: Arc::new(RwLock::new(system)),
            id: TypeId::of::<S>(),
        }
    }
}

pub struct EventHandlerHandle {
    pub system: Arc<RwLock<dyn Any + Send + Sync>>,
    pub event_id: TypeId,
    pub handler: Box<dyn EventHandlerWrapper>,
}

impl EventHandlerHandle {
    pub fn new(
        system: Arc<RwLock<dyn Any + Send + Sync>>,
        handler: Box<dyn EventHandlerWrapper>,
    ) -> Self {
        Self {
            system,
            event_id: handler.event_id(),
            handler,
        }
    }
}

pub trait EventHandlerBuilder {
    fn create(&self, handler: fn()) -> Box<dyn EventHandlerWrapper>;
}

pub struct TypedEventHandlerBuilder<S: System, T> {
    marker_system: PhantomData<S>,
    marker_event: PhantomData<T>,
}

impl<S: System, T: 'static> TypedEventHandlerBuilder<S, T> {
    pub fn get_builder() -> Box<dyn EventHandlerBuilder> {
        Box::new(Self {
            marker_system: PhantomData,
            marker_event: PhantomData,
        })
    }
}

impl<S: System, T: 'static> EventHandlerBuilder for TypedEventHandlerBuilder<S, T> {
    fn create(&self, handler: fn()) -> Box<dyn EventHandlerWrapper> {
        let handler: fn(&mut S, &T) = unsafe { std::mem::transmute(handler) };

        Box::new(RawEventHandler { handler })
    }
}

pub trait EventHandlerWrapper {
    fn event_id(&self) -> TypeId;

    fn handle(&self, system: &Arc<RwLock<dyn Any + Send + Sync>>, event: &dyn Any);
}

pub struct RawEventHandler<S, T> {
    handler: fn(&mut S, &T),
}

impl<S: System, T: 'static> EventHandlerWrapper for RawEventHandler<S, T> {
    fn event_id(&self) -> TypeId {
        TypeId::of::<T>()
    }

    fn handle(&self, system: &Arc<RwLock<dyn Any + Send + Sync>>, event: &dyn Any) {
        let mut system_guard = system.write();
        let system = system_guard
            .downcast_mut::<S>()
            .expect("RawEventHandler received incorrect system");
        let event = event
            .downcast_ref::<T>()
            .expect("RawEventHandler received incorrect event");
        (self.handler)(system, event);
    }
}

#[derive(Clone, Copy, Debug)]
pub struct ConstTypeId {
    id: fn() -> TypeId,
}

impl ConstTypeId {
    pub const fn of<T: 'static>() -> Self {
        Self {
            id: TypeId::of::<T>,
        }
    }
}

impl From<ConstTypeId> for TypeId {
    fn from(value: ConstTypeId) -> Self {
        (value.id)()
    }
}
